import argparse
from importlib import import_module

from .constants import COMMANDS


def main():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(required=True)

    for command_name in COMMANDS:
        command = import_module(f".{command_name}", "cici.cli")
        getattr(command, f"{command_name}_parser")(subparsers=subparsers)

    args = parser.parse_args()
    args.func(parser=parser, args=args)



ACCESS = "access"
ALWAYS = "always"
DEVELOPMENT = "development"
IF_NOT_PRESENT = "if-not-present"
NEVER = "never"
ON_SUCCESS = "on_success"
ON_FAILURE = "on_failure"
OTHER = "other"
PREPARE = "prepare"
PRODUCTION = "production"
PULL = "pull"
PULL_PUSH = "pull-push"
PUSH = "push"
STAGING = "staging"
START = "start"
STOP = "stop"
TESTING = "testing"
VERIFY = "verify"

CACHE_POLICIES = (PULL, PUSH, PULL_PUSH)

DEPLOYMENT_TIERS = (
    PRODUCTION,
    STAGING,
    TESTING,
    DEVELOPMENT,
    OTHER,
)

ENVIRONMENT_ACTIONS = (
    START,
    PREPARE,
    STOP,
    VERIFY,
    ACCESS,
)

PULL_POLICIES = (
    ALWAYS,
    IF_NOT_PRESENT,
    NEVER,
)

WHEN_CHOICES = (ON_SUCCESS, ON_FAILURE, ALWAYS)

RETRY_MAX_CHOICES = (0, 1, 2)

UNKNOWN_FAILURE = "unknown_failure"
SCRIPT_FAILURE = "script_failure"
API_FAILURE = "api_failure"
STUCK_OR_TIMEOUT_FAILURE = "stuck_or_timeout_failure"
RUNNER_SYSTEM_FAILURE = "runner_system_failure"
RUNNER_UNSUPPORTED = "runner_unsupported"
STALE_SCHEDULE = "stale_schedule"
JOB_EXECUTION_TIMEOUT = "job_execution_timeout"
ARCHIVED_FAILURE = "archived_failure"
UNMET_PREREQUISITES = "unmet_prerequisites"
SCHEDULER_FAILURE = "scheduler_failure"
DATA_INTEGRITY_FAILURE = "data_integrity_failure"

RETRY_WHEN_CHOICES = (
    ALWAYS,
    UNKNOWN_FAILURE,
    SCRIPT_FAILURE,
    API_FAILURE,
    STUCK_OR_TIMEOUT_FAILURE,
    RUNNER_SYSTEM_FAILURE,
    RUNNER_UNSUPPORTED,
    STALE_SCHEDULE,
    JOB_EXECUTION_TIMEOUT,
    ARCHIVED_FAILURE,
    UNMET_PREREQUISITES,
    SCHEDULER_FAILURE,
    DATA_INTEGRITY_FAILURE,
)

CI_FILE = ".gitlab-ci.yml"



from typing import Any, Optional, Union

import attrs
from attrs import define, field

from .constants import (
    ALWAYS,
    CACHE_POLICIES,
    DEPLOYMENT_TIERS,
    ENVIRONMENT_ACTIONS,
    ON_SUCCESS,
    PULL_PUSH,
    RETRY_MAX_CHOICES,
    START,
    WHEN_CHOICES,
)


@define(frozen=True, kw_only=True, slots=True)
class AllowFailure:
    exit_codes: Union[int, list[int]] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class CoverageReport:
    coverage_format: str
    path: str


@define(frozen=True, kw_only=True, slots=True)
class ArtifactReports:
    coverage_report: Optional[CoverageReport] = None
    junit: Union[str, list[str]] = field(factory=list)
    terraform: str = ""


@define(frozen=True, kw_only=True, slots=True)
class Artifacts:
    name: str = ""
    exclude: list[str] = field(factory=list)
    expire_in: str = ""
    expose_as: str = ""
    public: bool = True
    paths: list[str] = field(factory=list)
    reports: Optional[ArtifactReports] = None
    untracked: bool = False
    when: str = field(validator=attrs.validators.in_(WHEN_CHOICES), default=ON_SUCCESS)


@define(frozen=True, kw_only=True, slots=True)
class CacheKey:
    files: list[str] = field(factory=list)
    prefix: str = ""


@define(frozen=True, kw_only=True, slots=True)
class Cache:
    key: Union[str, CacheKey] = ""
    paths: list[str] = field(factory=list)
    untracked: bool = False
    unprotect: bool = False
    when: str = field(validator=attrs.validators.in_(WHEN_CHOICES), default=ON_SUCCESS)
    policy: str = field(
        validator=attrs.validators.in_(CACHE_POLICIES), default=PULL_PUSH
    )
    fallback_keys: list[str] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class EnvironmentKubernetes:
    namespace: str = ""


@define(frozen=True, kw_only=True, slots=True)
class Environment:
    name: str
    url: str = ""
    on_stop: str = ""
    action: str = field(
        validator=attrs.validators.in_(ENVIRONMENT_ACTIONS), default=START
    )
    auto_stop_in: str = ""
    kubernetes: Optional[EnvironmentKubernetes] = None
    deployment_tier: Optional[str] = field(
        validator=attrs.validators.in_(DEPLOYMENT_TIERS), default=None
    )


@define(frozen=True, kw_only=True, slots=True)
class Hooks:
    pre_get_sources_script: Union[str, list[Union[str, list[str]]]] = field(
        factory=list
    )


@define(frozen=True, kw_only=True, slots=True)
class IDToken:
    aud: Union[str, list[str]] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class Image:
    name: str
    entrypoint: Union[str, list[str]] = field(factory=list)
    pull_policy: Union[str, list[str]] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class IncludeLocal:
    local: str


@define(frozen=True, kw_only=True, slots=True)
class IncludeProject:
    project: str
    ref: Optional[str] = None
    file: Union[str, list[str]] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class IncludeRemote:
    remote: str


@define(frozen=True, kw_only=True, slots=True)
class IncludeTemplate:
    template: str


@define(frozen=True, kw_only=True, slots=True)
class RuleChanges:
    compare_to: str
    paths = list[str]


@define(frozen=True, kw_only=True, slots=True)
class Rule:
    if_: str = ""
    when: str = ""
    changes: Union[list[str], RuleChanges] = field(factory=list)
    exists: list[str] = field(factory=list)
    allow_failure: bool = False
    needs: Optional[Union[list[str], dict[str, str]]] = None
    variables: dict[str, str] = field(factory=dict)


@define(frozen=True, kw_only=True, slots=True)
class Retry:
    max: int = field(validator=attrs.validators.in_(RETRY_MAX_CHOICES), default=0)
    when: Union[str, list[str]] = ALWAYS


@define(frozen=True, kw_only=True, slots=True)
class Variable:
    description: Optional[str] = None
    value: Optional[str] = None
    options: Optional[list[str]] = None
    expand: bool = True


@define(frozen=True, kw_only=True, slots=True)
class Service:
    name: str
    entrypoint: Optional[Union[str, list[str]]] = None
    command: Optional[Union[str, list[str]]] = None
    variables: dict[str, Union[str, Variable]] = field(factory=dict)
    alias: Optional[str] = None
    pull_policy: Union[str, list[str]] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class Job:
    extends: Union[str, list[str]] = []
    stage: Optional[str] = None
    image: Optional[Union[str, Image]] = None
    services: list[Union[str, Service]] = field(factory=list)
    variables: dict[str, Union[str, Variable]] = field(factory=dict)
    before_script: Union[str, list[Union[str, list[str]]]] = field(factory=list)
    script: Union[str, list[Union[str, list[str]]]] = field(factory=list)
    after_script: Union[str, list[Union[str, list[str]]]] = field(factory=list)
    allow_failure: Union[bool, AllowFailure] = False
    artifacts: Optional[Artifacts] = None
    coverage: str = ""
    cache: Optional[Cache] = None
    dependencies: list[str] = field(factory=list)
    environment: Union[str, Environment] = ""
    hooks: Optional[Hooks] = None
    id_tokens: dict[str, IDToken] = field(factory=list)
    interruptible: bool = False
    retry: Union[int, Retry] = 0
    rules: list[Rule] = field(factory=list)
    tags: list[str] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class Workflow:
    name: str = ""
    rules: list[Rule] = field(factory=list)


@define(frozen=True, kw_only=True, slots=True)
class Default:
    after_script: Union[str, list[Union[str, list[str]]]] = field(factory=list)
    artifacts: Optional[Artifacts] = None
    before_script: Union[str, list[Union[str, list[str]]]] = field(factory=list)
    cache: Optional[Cache] = None
    hooks: Optional[Hooks] = None
    image: Optional[Union[str, Image]] = None
    interruptible: bool = False
    retry: Union[int, Retry] = 0
    tags: list[str] = field(factory=list)


@define(kw_only=True, slots=True)
class File:
    jobs: dict[str, Job] = field(factory=dict)
    stages: list[str] = field(factory=list)
    include: Union[
        str, list[Union[IncludeLocal, IncludeProject, IncludeRemote, IncludeTemplate]]
    ] = field(factory=list)
    workflow: Optional[Workflow] = None
    variables: dict[str, Union[str, Variable]] = field(factory=dict)
    extras: dict[str, Any] = field(factory=dict)



import copy
import logging
from importlib import import_module
from pathlib import Path
from typing import Optional

from termcolor import colored

from ..constants import CONFIG_DIR_NAME, DEFAULT_PROVIDER


def get_job_prefix(pipeline_name: str, job_name: str) -> Optional[str]:
    if job_name.startswith("."):
        return None
    return f"{pipeline_name}-"


def parse_groups(name, file, groups):
    if groups:
        groups = [group.strip() for group in groups.split(",")]

    if not groups:
        group_names = set()
        for job_name in file.jobs.keys():
            if job_name.startswith("."):
                continue
            group_name = job_name[len(f"{name}-") :]
            group_name = group_name.split("-")[0]
            group_names.add(group_name)
        groups = sorted(list(group_names))
    return groups


def bundle_command(parser, args):
    logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")
    print(colored("pipeline name:", color="yellow"), args.pipeline_name)

    provider = import_module(f".{DEFAULT_PROVIDER}", "cici.providers")

    ci_file_path = args.config_path / provider.CI_FILE

    if not Path(ci_file_path).exists():
        parser.error(f"file not found: {ci_file_path}")

    file = provider.load(ci_file_path)

    args.output_path = Path(args.output_path)
    args.groups = parse_groups(args.pipeline_name, file, args.groups)
    print(colored("bundle names:", color="yellow"), args.groups)

    for bundle_name in args.groups:
        pattern = f"{args.pipeline_name}-{bundle_name}"

        bundle = copy.deepcopy(file)
        bundle.jobs = {}

        for job_name, job in file.jobs.items():
            job_prefix = get_job_prefix(
                pipeline_name=args.pipeline_name,
                job_name=job_name,
            )
            if not job_prefix:
                continue
            assert job_name.startswith(job_prefix)
            if job_name.startswith(pattern):
                bundle.jobs[job_name] = job

        bundle_filename = args.output_path / f"{bundle_name}.yml"

        with open(bundle_filename, "w") as stream:
            provider.dump(bundle, stream)
        print(colored("created", "magenta"), bundle_filename.name)


def bundle_parser(subparsers):
    parser = subparsers.add_parser(
        "bundle", help="distill a CI file to a job and its dependencies"
    )
    parser.add_argument(
        "config_path",
        metavar="DIR",
        nargs="?",
        type=Path,
        default=(Path.cwd() / CONFIG_DIR_NAME).absolute(),
    )
    parser.add_argument("-g", "--groups", help="job group patterns")
    parser.add_argument(
        "-o",
        "--output",
        metavar="DIR",
        dest="output_path",
        type=Path,
        default=Path.cwd().absolute(),
    )
    parser.add_argument(
        "-n",
        "--name",
        metavar="DIR",
        dest="pipeline_name",
        default=str(Path.cwd().name),
        help="the name of the pipeline",
    )
    parser.set_defaults(func=bundle_command)
    return parser
